import os
import sys
import unittest
from pathlib import PurePosixPath
import pytest
from augpathlib import swap
from augpathlib import exceptions as exc
from augpathlib import AugmentedPath, LocalPath
from augpathlib import SymlinkCache, PrimaryCache
from augpathlib import PathMeta
from augpathlib import PathMeta
from augpathlib.meta import _PathMetaAsSymlink, _PathMetaAsXattrs
from .common import (log,
                     onerror,
                     project_path,
                     temp_path,
                     test_base,
                     test_path,
                     TestPathHelper,
                     LocalPathTest,
                     CachePathTest,)

SymlinkCache._local_class = AugmentedPath  # have to set a default

class Helper:
    @classmethod
    def setUpClass(cls):
        if not test_base.exists():
            test_base.mkdir()

    @classmethod
    def tearDownClass(cls):
        if test_base.exists():
            test_base.rmtree(onerror=onerror)

    def setUp(self):
        self.test_link = AugmentedPath(test_base, 'evil-symlink')  # FIXME random needed ...
        if self.test_link.is_symlink():
            self.test_link.unlink()

        self.test_link.symlink_to('hello/there')

        self.test_path = AugmentedPath(test_base, 'aug-testpath')  # FIXME random needed ...
        if self.test_path.exists():
            self.test_path.rmtree(onerror=onerror)

    def tearDown(self):
        if self.test_link.is_symlink():
            self.test_link.unlink()

        if self.test_path.exists():
            self.test_path.rmtree(onerror=onerror)


class TestAugPath(Helper, unittest.TestCase):

    def test_is_dir_symlink(self):
        assert not self.test_link.is_dir()

    def test_rmtree(self):
        self.test_path.mkdir()
        d = (self.test_path / 'heh')
        d.mkdir()
        f = (self.test_path / 'other')
        f.touch()
        f.chmod(0o0000)
        self.test_path.rmtree(onerror=onerror)

    def test_rmtree_ignore(self):
        try:
            self.test_path.rmtree(onerror=onerror)
            raise AssertionError('should fail')
        except FileNotFoundError as e:
            pass

        # this doesn't test passing deeper ...
        self.test_path.rmtree(ignore_errors=True, onerror=onerror)

    def test_rmtree_symlinks(self):
        """ make sure we don't rmtree through symlinks """
        tp = self.test_path
        tp.mkdir()
        d = tp / 'dir'
        t = tp / 'dir/target'
        f = tp / 'dir/target/file'
        so = tp / 'dir/source'
        s = tp / 'dir/source/symlink'
        d.mkdir()
        t.mkdir()
        f.touch()
        so.mkdir()
        s.symlink_to(t)
        try:
            s.rmtree()
            assert False, 'should have failed!'
        except OSError:
            pass

        so.rmtree()
        assert not so.exists()
        assert f.exists() and t.exists()
        self.test_path.rmtree(onerror=onerror)

    def test_relative_path_from(self):
        p1 = self.test_path / 'a' / 'b' / 'c' / 'd'
        p2 = self.test_path / 'e' / 'f' / 'g' / 'h'
        e1 = AugmentedPath('..', '..', '..', 'a', 'b', 'c', 'd')
        e2 = AugmentedPath('..', '..', '..', 'e', 'f', 'g', 'h')
        p1rfp2 = p1.relative_path_from(p2)
        p2rfp1 = p2.relative_path_from(p1)
        assert e1 == p1rfp2, p1rfp2
        assert e2 == p2rfp1, p2rfp1


@pytest.mark.skipif(os.name == 'nt', reason='no easy way to get libmagic on windows')
class TestMimetype(Helper, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.tf = self.test_path / 'some-text.txt'
        self.test_path.mkdir()
        with open(self.tf, 'wt') as f:
            f.write('hello')

    def test_mimetype(self):
        mt = self.tf.mimetype
        assert mt == 'text/plain', mt

    def test_magic_mimetype(self):
        mmt = self.tf._magic_mimetype
        assert mmt == 'text/plain', mmt


class TestAugPathCopy(Helper, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.test_path.mkdir()
        self.source_d = self.test_path / 'source-dir'
        self.target_d = self.test_path / 'target-dir'
        self.source_f = self.test_path / 'source-file'
        self.target_f = self.test_path / 'target-file'
        self.source_d.mkdir()
        self.target_d.mkdir()
        self.source_f.touch()
        self.target_f.touch()

    def test_copy_outto(self):
        #self.source_d.copy_outto(self.target_d)  # copytree not currently supported
        self.source_f.copy_outto(self.target_d)

    def test_copy_outto_fail_d(self):
        try:
            self.source_d.copy_outto(self.target_f)
            raise AssertionError('should have failed with NotADirectoryError')
        except NotADirectoryError:
            pass

    def test_copy_outto_fail_f(self):
        try:
            self.source_f.copy_outto(self.target_f)
            raise AssertionError('should have failed with NotADirectoryError')
        except NotADirectoryError:
            pass

    @pytest.mark.skip('copytree not implemented')
    def test_copy_outto_fail_existing_d(self):
        self.source_d.copy_outto(self.target_d)
        try:
            self.source_d.copy_outto(self.target_d)
            raise AssertionError('should have failed with FileExistsError')
        except FileExistsError:
            pass

    def test_copy_outto_fail_existing_f(self):
        self.source_f.copy_outto(self.target_d)
        try:
            self.source_f.copy_outto(self.target_d)
            raise AssertionError('should have failed with FileExistsError')
        except exc.PathExistsError:
            pass

    def test_copy_infrom(self):
        #self.target_d.copy_infrom(self.source_d)  # copytree not implemented
        self.target_d.copy_infrom(self.source_f)
    
    @pytest.mark.skip('copytree not implemented')
    def test_copy_infrom_fail_d(self):
        try:
            self.target_f.copy_infrom(self.source_d)
            raise AssertionError('should have failed with NotADirectoryError')
        except NotADirectoryError:
            pass

    def test_copy_infrom_fail_f(self):
        try:
            self.target_f.copy_infrom(self.source_f)
            raise AssertionError('should have failed with NotADirectoryError')
        except NotADirectoryError:
            pass


class TestAugPathSwap(Helper, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.test_path.mkdir()
        self.source_d = self.test_path / 'source-dir'
        self.target_d = self.test_path / 'target-dir'
        self.source_d.mkdir()
        self.target_d.mkdir()

        self.f1 = self.source_d / 'f1'
        self.f2 = self.target_d / 'f2'
        self.f1.touch()
        self.f2.touch()

    def _doit(self, thunk):
        assert (self.source_d / self.f1.name).exists()
        assert (self.target_d / self.f2.name).exists()
        assert self.f1.exists()
        assert self.f2.exists()
        thunk()
        assert (self.source_d / self.f2.name).exists()
        assert (self.target_d / self.f1.name).exists()
        assert not self.f1.exists()
        assert not self.f2.exists()

    @pytest.mark.skipif(sys.platform != 'linux', reason='not implemented')
    def test_swap_linux(self):
        self._doit(lambda :swap.swap_linux(self.source_d, self.target_d))

    @pytest.mark.skipif(sys.platform != 'linux', reason='not implemented')
    def test_swap(self):
        self._doit(lambda :self.source_d.swap(self.target_d))

    def test_swap_carefree(self):
        self._doit(lambda :self.source_d.swap_carefree(self.target_d))


class TestACachePath(unittest.TestCase):
    def setUp(self):
        if test_path.is_symlink():
            test_path.unlink()

        if test_path.exists() and test_path.is_dir():
            test_path.rmdir()

    def test_0_exists(self):
        log.debug(CachePathTest.anchor)
        assert not CachePathTest.anchor.is_symlink()
        assert CachePathTest.anchor.meta is None

    def test_1_create(self):
        wat = CachePathTest(test_path, meta=PathMeta(id='0'))
        log.debug(wat)
        assert wat.meta

    def test_2_create_dir(self):
        test_path.mkdir()
        wat = CachePathTest(test_path, meta=PathMeta(id='0'))
        log.debug(wat)
        assert wat.meta

    def test_cache_init_dir(self):
        if test_path.exists():  # in case something went wrong with a previous test
            shutil.rmtree(test_path, onerror=onerror)

        test_path.mkdir()
        assert test_path.exists()
        assert test_path.is_dir()
        test_path.cache_init('0')
        assert test_path.cache
        assert test_path.cache.meta


class TestCacheSparse(unittest.TestCase):
    _test_class = CachePathTest
    sandbox = test_base / 'sparse-sandbox'
    @classmethod
    def setUpClass(cls):
        cls.sandbox.mkdir(parents=True)

    @classmethod
    def tearDownClass(cls):
        cls.sandbox.rmtree(onerror=onerror)

    def setUp(self):
        self.dir = self._test_class(self.sandbox, 'some-dir', meta=PathMeta(id='0'))
        if self.dir.exists():
            self.dir.rmtree(onerror=onerror)
        elif self.dir.is_symlink():
            self.dir.unlink()
        self.dir.mkdir()

        self.file = self._test_class(self.sandbox, 'some-file', meta=PathMeta(id='1'))
        if self.file.exists() or self.file.is_symlink():
            self.file.unlink()
        self.file.touch()

    def tearDown(self):
        self.dir.rmtree(onerror=onerror)
        self.file.unlink()

    def test_sparse(self):
        test_file = self._test_class(self.dir, 'more-test', meta=PathMeta(id='3'))
        test_file.unlink()  # FIXME fix bad Cache constructor behavior already
        test_file.touch()

        assert not self.file.is_sparse()
        assert not self.dir.is_sparse()
        assert not test_file.is_sparse()

        self.dir._mark_sparse()
        assert not self.file.is_sparse()
        assert self.dir.is_sparse()
        assert test_file.is_sparse()

        self.dir._clear_sparse()
        assert not self.file.is_sparse()
        assert not self.dir.is_sparse()
        assert not test_file.is_sparse()


class TestPathMeta(unittest.TestCase):
    prefix = None

    def setUp(self):
        self.path = LocalPathTest(project_path)

        self.test_path = LocalPathTest(test_base, 'testpath')  # FIXME random needed ...
        if self.test_path.is_symlink():
            self.test_path.unlink()

    def _test_getattr_size_hr(self):
        pm = PathMeta(size=1000000000000000)
        woo = getattr(pm, 'size.hr')

    def test_neg__neg__(self):
        pm = PathMeta(id='lol')
        assert pm

    def test___neg__(self):
        pm = PathMeta()
        assert not pm, set(pm.__dict__.values())

    def test_xattrs_roundtrip(self):
        # TODO __kwargs ...
        pm = self.path.meta
        xattrs = pm.as_xattrs(self.prefix)
        log.debug(xattrs)
        # FIXME actually write these to disk as well?
        new_pm = PathMeta.from_xattrs(xattrs, self.prefix)
        msg = '\n'.join([f'{k!r} {v!r} {getattr(new_pm, k)!r}' for k, v in pm.items()])
        assert new_pm == pm, msg
        #'\n'.join([str((getattr(pm, field), getattr(new_pm, field)))
        #for field in _PathMetaAsXattrs.fields])

    def test_metastore_roundtrip(self):
        pm = self.path.meta
        ms = pm.as_metastore(self.prefix)
        # FIXME actually write these to disk as well?
        new_pm = PathMeta.from_metastore(ms, self.prefix)
        assert new_pm == pm, '\n'.join([str((getattr(pm, field), getattr(new_pm, field)))
                                        for field in tuple()])  # TODO

    def test_symlink_roundtrip(self):
        meta = PathMeta(id='N:helloworld:123', size=10, checksum=b'1;o2j\x9912\xffo3ij\x01123,asdf.')
        path = self.test_path
        path._cache = SymlinkCache(path, meta=meta)
        path.cache.meta = meta
        new_meta = path.cache.meta
        path.unlink()
        msg = '\n'.join([f'{k!r} {v!r} {getattr(new_meta, k)!r}' for k, v in meta.items()])
        assert meta == new_meta, msg

    def _test_symlink_roundtrip_weird(self):
        path = LocalPathTest(test_base, 'testpath')  # FIXME random needed ...
        meta = PathMeta(id='N:helloworld:123', size=10, checksum=b'1;o2j\x9912\xffo3ij\x01123,asdf.')
        pure_symlink = PurePosixPath(path.name) / meta.as_symlink()
        path.symlink_to(pure_symlink)
        try:
            cache = SymlinkCache(path)
            new_meta = cache.meta
            msg = '\n'.join([f'{k!r} {v!r} {getattr(new_meta, k)!r}' for k, v in meta.items()])
            assert meta == new_meta, msg
        finally:
            path.unlink()

    def test_parts_roundtrip(self):
        pmas = _PathMetaAsSymlink()
        lpm = self.path.meta
        bpm = PathMeta(id='N:helloworld:123', size=10, checksum=b'1;o2j\x9912\xffo3ij\x01123,asdf.')
        bads = []
        for pm in (lpm, bpm):
            symlink = pm.as_symlink()
            log.debug(symlink)
            new_pm = pmas.from_parts(symlink.parts)
            #corrected_new_pm = PurePosixPath()
            if new_pm != pm:
                bads += ['\n'.join([str((getattr(pm, field), getattr(new_pm, field)))
                                    for field in ('id',) + _PathMetaAsSymlink.order
                                    if not (getattr(pm, field) is getattr(new_pm, field) is None)]),
                         f'{pm.__reduce__()}\n{new_pm.__reduce__()}']

        assert not bads, '\n===========\n'.join(bads)


class TestPrefix(TestPathMeta):
    prefix = 'prefix'


class TestPrefixEvil(TestPathMeta):
    prefix = 'prefix.'


class TestContext(unittest.TestCase):
    def setUp(self):
        if not temp_path.exists():
            temp_path.mkdir()

    def tearDown(self):
        temp_path.rmtree(onerror=onerror)

    def test_context(self):
        start = AugmentedPath.cwd()
        target = AugmentedPath(temp_path).resolve()  # resolve needed for osx
        distractor = AugmentedPath('~/').expanduser()
        assert temp_path.is_dir()
        with target:
            target_cwd = AugmentedPath.cwd()
            distractor.chdir()
            distractor_cwd = AugmentedPath.cwd()

        end = AugmentedPath.cwd()
        assert target == target_cwd, 'with target: failed'
        assert distractor == distractor_cwd, 'distractor cwd failed'
        assert start == end, 'it would seem that the distractor got us'
        assert start != target != distractor


class TestIdZero(TestPathHelper, unittest.TestCase):
    def test(self):
        zt = LocalPathTest(test_path) / 'zero-test'
        cache = CachePathTest(zt, meta=PathMeta(id='0'))
        assert cache.meta


class TestUpdateMeta(unittest.TestCase):
    def test_update(self):
        old = PathMeta(id='0', size=10, file_id=1)
        new = PathMeta(id='0', size=10, checksum='asdf')
        changed, merged = PrimaryCache._update_meta(old, new)
        test_value = PathMeta(id='0', size=10, file_id=1, checksum='asdf')
        assert merged == test_value, test_value.as_pretty_diff(merged)


class TestActuallyLocalPath(unittest.TestCase):
    def setUp(self):
        LocalPath

    def tearDown(self):
        LocalPath

    def test_no_cache_class(self):
        lp = LocalPath(__file__)
        p = lp.parent
        rc = list(p.rchildren)
        assert rc, 'hrm'



import os
import unittest
import pytest
import augpathlib as aug
from .common import test_base, onerror

sandbox = test_base / 'eat-sandbox'


class EatPath(aug.EatPath, aug.AugmentedPath): pass
EatPath._bind_flavours()


class EatXopenPath(aug.EatPath, aug.XopenPath, aug.AugmentedPath): pass
EatXopenPath._bind_flavours()


class TestEat(unittest.TestCase):
    _test_class = EatPath
    @classmethod
    def setUpClass(cls):
        sandbox.mkdir(parents=True)

    @classmethod
    def tearDownClass(cls):
        sandbox.rmtree(onerror=onerror)
    
    def setUp(self):
        self.dir = self._test_class(sandbox, 'some-dir')
        if self.dir.exists():
            self.dir.rmtree(onerror=onerror)
        self.dir.mkdir()

        self.file = self._test_class(sandbox, 'some-file')
        if self.file.exists():
            self.file.unlink()
        self.file.touch()

    def tearDown(self):
        self.dir.rmtree(onerror=onerror)
        self.file.unlink()

    @pytest.mark.skipif(os.name != 'nt', reason='This ADS behavior is windows only')
    def test_dir_simple(self):
        stream = self.dir._stream('wat')
        tv = b'wat-value'
        with open(stream, 'wb') as f:
            f.write(tv)

        with open(stream, 'rb') as f:
            test = f.read()

        assert test == tv
        streams = list(self.dir._streams)
        assert streams

    def test_dir(self):
        self.dir.setxattr('key', b'value')
        test = self.dir.xattrs()
        assert test

    def test_file(self):
        self.file.setxattr('key', b'value')
        test = self.file.xattrs()
        assert test

    def test_sparse(self):
        test_file = self.dir / 'more-test'
        test_file.touch()

        assert not self.file.is_sparse()
        assert not self.dir.is_sparse()
        assert not test_file.is_sparse()

        self.dir._mark_sparse()
        assert not self.file.is_sparse()
        assert self.dir.is_sparse()
        assert test_file.is_sparse()

        self.dir._clear_sparse()
        assert not self.file.is_sparse()
        assert not self.dir.is_sparse()
        assert not test_file.is_sparse()


class TestEatXopen(TestEat):
    _test_class = EatXopenPath



import os
import sys
import atexit
import pathlib
import warnings
import subprocess
from augpathlib import exceptions as exc
from augpathlib.meta import PathMeta
from augpathlib import caches, LocalPath
from augpathlib.utils import _bind_sysid_, StatResult, cypher_command_lookup, log
if os.name != 'nt':
    # pexpect on windows does not support pxssh
    # because it is missing spawn
    from pexpect import pxssh


class RemotePath:
    """ Remote data about a remote object. """

    _cache_class = None
    _debug = False

    # ugh this is such a bad implementation, let the remote paths exists
    # and init, and then just check if they exist, a path is not an object
    # that always dereferences ... what the heck was I thinking when I did this ...

    # we use a PurePath becuase we still want to key off this being local path
    # but we don't want any of the local file system operations to work by accident
    # so for example self.stat should return the remote value not the local value
    # which is what would happen if we used a PosixPath as the base class

    # need a way to pass the session information in independent of the actual path
    # abstractly having remote.data(global_id_for_local, self)
    # should be more than enough, the path object shouldn't need
    # to know that it has a remote id, the remote manager should
    # know that

    @classmethod
    def _new(cls, local_class, cache_class):
        """ when constructing a new remote using _new you MUST
            call init afterward to bind the remote api """

        # FIXME 1:1ness issue from local -> cache
        # probably best to force the type of the cache
        # to switch if there are multiple remote mappings
        # since there can be only 1 local file with the same
        # path, a composite cache or a multi-remote cache
        # seems a bit saner, or require explicit switching of
        # the active remote if one-at-a-time semantics are desired

        newcls = type(cls.__name__,
                      (cls,),
                      dict(_local_class=local_class,
                           _cache_class=cache_class))

        local_class._remote_class = newcls
        local_class._cache_class = cache_class
        cache_class._remote_class = newcls
        cache_class._local_class = local_class

        newcls.weighAnchor()
        cache_class.weighAnchor()

        return newcls

    @classmethod
    def init(cls, identifier):
        """ initialize the api from an identifier and bind the root """
        if not hasattr(cls, '_api'):
            cls._api = cls._api_class(identifier)
            cls.root = cls._api.root

        else:
            raise ValueError(f'{cls} already bound an api to {cls._api}')

    @classmethod
    def anchorToCache(cls, cache_anchor, init=True):
        # FIXME need to check for anchor after init and init after anchor
        if not hasattr(cls, '_cache_anchor'):
            if init:
                if not hasattr(cls, '_api'):
                    cls.init(cache_anchor.id)

            if hasattr(cls, 'root') and cls.root != cache_anchor.id:
                raise ValueError('root and anchor ids do not match! '
                                 f'{cls.root} != {cache_anchor.id}')

            cls._cache_anchor = cache_anchor
            return cls._cache_anchor
        else:
            raise ValueError(f'already anchored to {cls._cache_anchor}')

    @classmethod
    def anchorTo(cls, path, create=False):
        """ You already know the rock you want and
            you want the anchor stuck to it. """
        # FIXME should we fail on create=True and exists?

        if isinstance(path, caches.CachePath):
            # FIXME the non-existence problem rears its head again
            return cls.anchorToCache(path)
        elif isinstance(path, LocalPath):
            # FIXME the non-existence problem rears its head again
            if path.cache:
                return cls.anchorToCache(path.cache)
            else:
                root = cls.root if isinstance(cls.root, cls) else cls(cls.root)
                if path.name != root.name:
                    # unlike git you cannot clone to a folder with a different
                    # name (for now ... maybe can figure out how in the future)
                    raise ValueError('Path name and root name do not match.'
                                     f'{path.name} != {cls.root.name}')

                if create:
                    return cls.dropAnchor(path.parent)  # existing folder dealt with in dropAnchor
                else:
                    raise ValueError(f'not creating {path} since create=False')
        else:
            raise TypeError(f"Don't know how to anchor to a {type(path)} {path}")

    @classmethod
    def _get_local_root_path(cls, parent_path=None):
        if parent_path is None:
            parent_path = cls._local_class.cwd()
        else:
            parent_path = cls._local_class(parent_path)

        root = cls(cls.root)  # FIXME formalize the use of root
        path = parent_path / root.name
        return root, path

    @classmethod
    def smartAnchor(cls, parent_path=None):
        # work around the suspect logic
        # in the implementation below
        try:
            return cls.dropAnchor(parent_path=parent_path)
        except exc.RemoteAlreadyAnchoredError as e:
            root, path = cls._get_local_root_path(parent_path)
            if cls._cache_anchor == path.cache:
                return cls._cache_anchor
            else:
                raise e  # possibly check if the anchor is the same?

        except exc.CacheExistsError as e:
            root, path = cls._get_local_root_path(parent_path)
            cls._cache_anchor = path.cache
            return cls._cache_anchor

        except exc.DirectoryNotEmptyError as e:
            root, path = cls._get_local_root_path(parent_path)
            if path.cache:
                cls._cache_anchor = path.cache
                return cls._cache_anchor
            else:
                raise e

    @classmethod
    def dropAnchor(cls, parent_path=None):
        """ If a _cache_anchor does not exist then create it,
            otherwise raise an error. If a local anchor already
            exists do not use this method.

            You know that the ship (path) is more or less in the right
            place but you don't know for sure exactly which rock the
            anchor will catch on (you don't know the name of the remote).
        """
        if not hasattr(cls, '_cache_anchor'):
            root, path = cls._get_local_root_path(parent_path)
            if not path.exists():
                if root.is_file():
                    raise NotImplementedError(
                        'Have not implemented mapping for individual files yet.')

                elif root.is_dir():
                    path.mkdir()

                else:
                    raise NotImplementedError(f'What\'s a {root}?!')

            elif list(path.children):
                raise exc.DirectoryNotEmptyError(f'has children {path}')

            cls._cache_anchor = path.cache_init(root.id, anchor=True)
            # we explicitly do not handle the possible CacheExistsError here
            # so that there is a path where anchoring can fail loudly
            # we may not need that at the end of the day, but we will see
            return cls._cache_anchor

        else:
            raise exc.RemoteAlreadyAnchoredError(f'{cls} already anchored to '
                                                 f'{cls._cache_anchor}')

    @classmethod
    def weighAnchor(cls):
        # TODO determine whether the current behavior is correct
        # calling this will not cause the cache class to weigh anchor
        # but there is a small chance that it should

        # TODO is _abstract_class needed here? or do we not need it
        # because remote paths don't have the crazy hierarchy that
        # pathlib derived paths do? and will this change when we fix
        # everything ...

        if hasattr(cls, '_cache_anchor'):
            delattr(cls, '_cache_anchor')

    @classmethod
    def setup(cls, local_class, cache_class):
        """ call this once to bind everything together """

        cn = self.__class__.__name__
        warnings.warn(f'{cn}.setup is deprecated please switch to RemotePath._new',
                      DeprecationWarning,
                      stacklevel=2)

        cache_class.setup(local_class, cls)

    def bootstrap(self, recursive=False, only=tuple(), skip=tuple(), sparse=tuple()):
        #self.cache.remote = self  # duh
        # if you forget to tell the cache you exist of course it will go to
        # the internet to look for you, it isn't quite smart enough and
        # we're trying not to throw dicts around willy nilly here ...
        return self.cache.bootstrap(self.meta, recursive=recursive, only=only, skip=skip, sparse=sparse)

    def __init__(self, thing_with_id, cache=None):
        if isinstance(thing_with_id, str):
            id = thing_with_id
        elif isinstance(thing_with_id, PathMeta):
            id = thing_with_id.id
        elif isinstance(thing_with_id, RemotePath):
            id = thing_with_id.id
        else:
            raise TypeError(f'Don\'t know how to initialize a remote from {thing_with_id}')

        self._id = id
        if cache is not None:
            self._cache = cache
            self.cache._remote = self

        self._errors = []

    @property
    def id(self):
        return self._id

    @property
    def errors(self):
        raise NotImplementedError

    @property
    def cache(self):
        if hasattr(self, '_cache_anchor') and self._cache_anchor is not None:
            return self._cache
        else:
            # cache is not real
            class NullCache:
                @property
                def local(self, remote=self):
                    raise TypeError(f'No cache for {remote}')

                @property
                def _are_we_there_yet(self, remote=self):
                    # this is useless since these classes are ephemoral
                    if hasattr(remote, '_cache_anchor') and remote._cache_anchor is not None:
                        remote.cache_init()

                def __rtruediv__(self, other):
                    return None

                def __truediv__(self, other):
                    return None

            return NullCache()

    def cache_init(self, parents=False):
        try:
            return self._cache_anchor / self
        except FileNotFoundError:
            if parents:
                #parent, *rest = self.parent.cache_init(parents=parents)
                #return (self.cache_init(), parent, *rest)
                parent = self.parent
                parent_cache = parent.cache_init(parents=parents)
                parent_cache.local.cache_init(parent.meta) # FIXME hrm we shouldn't have to do this
                # and it isn't working anyway ... the xattrs don't seem to be getting set
                return self.cache_init()
            else:
                raise

    @property
    def _cache(self):
        """ To catch a bad call to set ... """
        if hasattr(self, '_c_cache'):
            return self._c_cache

    @_cache.setter
    def _cache(self, cache):
        if not isinstance(cache, caches.CachePath):
            raise TypeError(f'cache is a {type(cache)} not a CachePath!')
        #elif cache.meta is None:  # useful for certain debugging situations
            #raise ValueError(f'cache has no meta {cache}')

        self._c_cache = cache

    def _cache_setter(self, cache, update_meta=True):
        cache._remote = self
        # FIXME in principle
        # setting cache needs to come before update_meta
        # in the event that self.meta is missing file_id
        # if meta updater fails we unset self._c_cache
        self._cache = cache
        if update_meta:
            try:
                cache._meta_updater(self.meta)
            except BaseException as e:
                self._c_cache = None
                delattr(self, '_c_cache')
                raise e

    @property
    def parent_id(self):
        """ BEWARE if self.parent hits the network so will this.
            In the event that it does, overwrite this method. """

        return self.parent.id

    def _parent_changed(self, cache):
        return self.parent_id != cache.parent.id

    def _on_cache_move_error(self, error, cache):
        """ called after a failure to move a cached file to a new location """
        raise error

    def update_cache(self, cache=None, fetch=True):
        """ Update a cache object using the metadata attached to this remote.

            This is different form _cache_setter in that it runs update_meta
            by default, handles many more edge cases, and checks for consistency.
            _cache_setter is usually invoked internally by a CachePath method that
            wants to register itself with a remote as an implementaiton detail. """

        if cache is not None and self.cache is not None:
            # TODO see if there are any exceptions to this behavior
            raise TypeError('cannot accept cache kwarg when self.cache not None')
        elif cache is None:
            cache = self.cache

        parent_changed = self._parent_changed(cache)

        if self.cache is None:
            # HACK test if cache is not None before it may have been reassigned
            if cache.name != self.name:
                msg = ('Cannot update the name and content of a file at the '
                       'same time.\nAre you sure you have passed the right '
                       f'cache object?\n{cache.name} != {self.name}')
                raise ValueError(msg)

            elif parent_changed:
                msg = ('Cannot update the parent and content of a file at the '
                       'same time.\nAre you sure you have passed the right '
                       f'cache object?\n{cache.parent.id} != {self.parent_id}')
                raise ValueError(msg)

        log.debug(f'maybe updating cache for {self.name}')
        file_is_different = cache._meta_updater(self.meta, fetch=fetch)
        # update the cache first  # FIXME this may be out of order ...
        # then move to the new name if relevant
        # prevents moving partial metadata onto existing files
        if cache.name != self.name or parent_changed:  # this is localy correct
            # the issue is that move is now smarter
            # and will detect if a parent path has changed
            try:
                cache.move(remote=self)
            except exc.WhyDidntThisGetMovedBeforeError as e:
                # AAAAAAAAAAAAAAAAAAAAAAAAAAAAA
                # deal with the sadness that is non-unique filenames
                # I am 99.999999999999999% certain that users do not
                # expect this behavior ...
                log.error(e)
                self._on_cache_move_error(e, cache)

        return file_is_different

    @property
    def local(self):
        return self.cache.local  # FIXME there are use cases for bypassing the cache ...

    @property
    def local_direct(self):
        # kind of uninstrumeted ???
        return self._local_class(self.as_path())

    @property
    def anchor(self):
        """ the semantics of anchor for remote paths are a bit different
            RemotePath code expects this function to return a RemotePath
            NOT a string as is the case for core pathlib. """
        raise NotImplementedError

    @property
    def _meta(self):  # catch stragglers
        raise NotImplementedError

    def refresh(self):
        """ Refresh the local in memory metadata for this remote.
            Implement actual functionality in your subclass. """

        raise NotImplementedError
        # could be fetch or pull, but there are really multiple pulls as we know

        # clear the cached value for _meta
        if hasattr(self, '_meta'):
            delattr(self, '_meta')

    @property
    def data(self):
        raise NotImplementedError
        self.cache.id
        for chunk in chunks:
            yield chunk

    @property
    def meta(self):
        # on blackfynn this is the package id or object id
        # this will error if there is no implementaiton if self.id
        raise NotImplementedError
        #return PathMeta(id=self.id)

    def _meta_setter(self, value):
        raise NotImplementedError

    @property
    def annotations(self):
        # these are models etc in blackfynn
        yield from []
        raise NotImplementedError

    def as_path(self):
        """ returns the relative path construction for the child so that local can make use of it """
        return pathlib.PurePath(*self.parts)

    def _parts_relative_to(self, remote, cache_parent=None):
        parent_names = []  # FIXME massive inefficient due to retreading subpaths :/
        # have a look at how pathlib implements parents
        parent = self.parent
        if parent != remote:
            parent_names.append(parent.name)
            # FIXME can this go stale? if so how?
            #log.debug(cache_parent)
            if cache_parent is not None and parent.id == cache_parent.id:
                    for c_parent in cache_parent.parents:
                        if c_parent is None:
                            continue
                        elif c_parent.name == remote.name:  # FIXME trick to avoid calling id
                            parent_names.append(c_parent.name)  # since be compare one earlier we add here
                            break
                        else:
                            parent_names.append(c_parent.name)

            else:
                for parent in parent.parents:
                    if parent == remote:
                        break
                    elif parent is None:
                        continue  # value error incoming
                    else:
                        parent_names.append(parent.name)

                else:
                    self._errors += ['file-deleted']
                    msg = f'{remote} is not one of {self}\'s parents'
                    log.error(msg)
                    #raise ValueError()

            args = (*reversed(parent_names), self.name)
        elif self == parent:
            args = ('',)
        else:
            args = self.name,

        return args

    @property
    def parts(self):
        if self == self.anchor:
            return tuple()

        if not hasattr(self, '_parts'):
            if self.cache:
                cache_parent = self.cache.parent
            else:
                cache_parent = None

            self._parts = tuple(self._parts_relative_to(self.anchor, cache_parent))

        return self._parts

    @property
    def parent(self):
        """ The atomic parent operation as understood by the remote. """
        raise NotImplementedError

    @property
    def parents(self):
        parent = self.parent
        while parent:
            yield parent
            parent = parent.parent

    @property
    def children(self):
        # uniform interface for retrieving remote hierarchies decoupled from meta
        raise NotImplementedError

    @property
    def rchildren(self):
        # uniform interface for retrieving remote hierarchies decoupled from meta
        yield from self._rchildren()

    def _rchildren(self, create_cache=True, sparse=False):
        raise NotImplementedError

    def children_pull(self, existing):
        # uniform interface for asking the remote to
        # update children using its own implementation
        raise NotImplementedError

    def iterdir(self):
        # I'm guessing most remotes don't support this
        raise NotImplementedError

    def glob(self, pattern):
        raise NotImplementedError

    def rglob(self, pattern):
        raise NotImplementedError

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return f'{self.__class__.__name__}({self.id!r})'


class SshRemote(RemotePath, pathlib.PurePath):
    """ Testing. To be used with ssh-agent.
        StuFiS The stupid file sync. """

    _cache_class = None  # set when calling __new__
    encoding = 'utf-8'

    _meta = None  # override RemotePath dragnet
    _meta_maker = LocalPath._meta_maker

    sysid = None
    _bind_sysid = classmethod(_bind_sysid_)

    @classmethod
    def _new(cls, local_class, cache_class):
        newcls = super()._new(local_class, cache_class)

        # must run before we can get the sysid, which is a bit odd
        # given that we don't actually sandbox the filesystem

        newcls._bind_flavours()
        return newcls

    @classmethod
    def _bind_flavours(cls, pos_helpers=tuple(), win_helpers=tuple()):
        pos, win = cls._get_flavours()

        if pos is None:
            pos = type(f'{cls.__name__}Posix',
                       (*pos_helpers, cls, pathlib.PurePosixPath), {})

        if win is None:
            win = type(f'{cls.__name__}Windows',
                       (*win_helpers, cls, pathlib.PureWindowsPath), {})

        cls.__abstractpath = cls
        cls.__posixpath = pos
        cls.__windowspath = win

    @classmethod
    def _get_flavours(cls):
        pos, win = None, None
        for subcls in cls.__subclasses__():  # direct only
            if subcls._flavour is pathlib._posix_flavour:
                pos = subcls
            elif subcls._flavour is pathlib._windows_flavour:
                win = subcls
            else:
                raise TypeError(f'unknown flavour for {cls} {cls._flavour}')

        return pos, win

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_flavour'):
            cls = cls.__windowspath if os.name == 'nt' else cls.__posixpath

        if isinstance(args[0], str) and args[0].startswith(cls.host + ':'):
            # FIXME not great but allows less verbose where possible ...
            # also possibly an opportunity to check if hostnames match?
            # ugh unix everything is a stream of bytes is annoying here
            _, *args = (args[0].split(':', 1), *args[1:])

        _self = pathlib.PurePath.__new__(cls, *args)  # no kwargs since the only kwargs are for init
        _self.remote_platform = _self._remote_platform
        return _self
    
        # TODO this isn't quite working yet due to bootstrapping issues as usual
        # it also isn't working because we want access to all paths in many cases
        # the root remains and the calculation against anchor remains for any
        # relative path that is provided, and the default behavior for absolute
        # paths protects us from sillyness

        if _self.id != cls.root: #_cache_anchor.id:
            self = _self.relative_to(_self.anchor)
        else:
            self = pathlib.PurePath.__new__(cls, '.')  # FIXME make sure this is interpreted correctly ...

        self._errors = []
        return self

    @classmethod
    def init(cls, host_path):
        """ should only be invoked after _new has bound local and cache classes """
        if not hasattr(cls, '_anchor'):
            cls.root = host_path  # I think this is right ...
            host, path = host_path.split(':', 1)

            if not hasattr(cls, '_flavour'):
                cls = cls.__windowspath if os.name == 'nt' else cls.__posixpath

            cls._anchor = pathlib.PurePath.__new__(cls, path)

            session = pxssh.pxssh(options=dict(IdentityAgent=os.environ.get('SSH_AUTH_SOCK')))
            session.login(host, ssh_config=LocalPath('~/.ssh/config').expanduser().as_posix())
            cls._rows = 200
            cls._cols = 200
            session.setwinsize(cls._rows, cls._cols)  # prevent linewraps of long commands
            session.prompt()
            atexit.register(lambda:(session.sendeof(), session.close()))
            cls.host = host
            cls.session = session
            cls._uid, *cls._gids = [int(i) for i in (cls._ssh('echo $(id -u) $(id -G)')
                                                    .decode().split(' '))]
        else:
            raise ValueError(f'{cls} already bound an remote to {cls._anchor}')

    @classmethod
    def anchorToCache(cls, cache_anchor, init=True):
        anchor = super().anchorToCache(cache_anchor=cache_anchor, init=init)
        # _cache_anchor has to be bound for _bind_sysid to work
        # that binding happens after init so we do this here
        cls._bind_sysid()
        return anchor

    def __init__(self, thing_with_id, cache=None):
        if isinstance(thing_with_id, pathlib.PurePath):
            thing_with_id = thing_with_id.as_posix()

        super().__init__(thing_with_id, cache=cache)

    @property
    def anchor(self):
        return self._anchor
        #return self._cache_anchor.remote
        # FIXME warning on relative paths ...
        # also ... might be convenient to allow
        # setting non-/ anchors, but perhaps for another day
        #return self.__class__('/', host=self.host)

    @property
    def id(self):
        return f'{self.host}:{self.rpath}'
        #return self.host + ':' + self.as_posix()  # FIXME relative to anchor?

    @property
    def cache_key(self):
        """ since some systems have compound ids ... """
        raise NotImplementedError

    @property
    def rpath(self):
        # FIXME relative paths when the anchor is set differently
        # the anchor will have to be stored as well since there coulde
        # be many possible anchors per host, thus, if an anchor relative
        # identifier is supplied then we need to construct the full path

        # conveniently in this case if self is a fully rooted path then
        # it will overwrite the anchor path
        # TODO make sure that the common path is the anchor ...
        return (self.anchor / self).as_posix()

    def _parts_relative_to(self, remote, cache_parent=None):
        if remote == self.anchor:
            # have to build from self.anchor._parts because it is the only
            # place the keeps the original parts
            remote = pathlib.PurePath(*self.anchor._parts)

        return self.relative_to(remote).parts

    def refresh(self):
        # TODO probably not the best idea ...
        raise NotImplementedError('This baby goes to the network every single time!')

    def access(self, mode):
        """ types are 'read', 'write', and 'execute' """
        try:
            st = self.stat()

        except (PermissionError, FileNotFoundError) as e:
            return False

        r, w, x = 0x124, 0x92, 0x49
        read    = ((r & st.st_mode) >> 2) & (mode == 'read'    or mode == os.R_OK) * x
        write   = ((w & st.st_mode) >> 1) & (mode == 'write'   or mode == os.W_OK) * x
        execute =  (x & st.st_mode)       & (mode == 'execute' or mode == os.X_OK) * x
        current = read + write + execute

        u, g, e = 0x40, 0x8, 0x1
        return (u & current and st.st_uid == self._uid or
                g & current and st.st_gid in self._gids or
                e & current)

    def open(self, mode='wt', buffering=-1, encoding=None,
             errors=None, newline=None):
        if mode not in ('wb', 'wt'):
            raise TypeError('only w[bt] mode is supported')  # TODO ...

        #breakpoint()
        return
        class Hrm:
            session = self.session
            def write(self, value):
                self.session

        #cmd = ['ssh', self.host, f'"cat - > {self.rpath}"']
        #self.session
        #p = subprocess.Popen()

    @property
    def data(self):
        cmd = ['scp', self.id, '/dev/stdout']
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        while True:
            data = p.stdout.read(4096)  # TODO hinting
            if not data:
                break

            yield data

        p.communicate()

    # reuse meta from local
    # def meta (make it easier to search for this)
    meta = LocalPath.meta  # magic

    #def _ssh(self, remote_cmd):
    @classmethod
    def _ssh(cls, remote_cmd):
        #print(remote_cmd)
        if len(remote_cmd) > cls._cols:
            raise exc.CommandTooLongError
        n_bytes = cls.session.sendline(remote_cmd)
        cls.session.prompt()
        raw = cls.session.before
        out = raw[n_bytes + 1:].strip()  # strip once here since we always will
        #print(raw)
        #print(out)
        return out

    @property
    def _remote_platform(self):
        remote_cmd = "uname -a | awk '{ print tolower($1) }'"
        return self._ssh(remote_cmd).decode(self.encoding)

    @property
    def cypher_command(self):
        # this one is a little backwards, because we can control
        # whatever cypher we want, unlike in other cases
        return cypher_command_lookup[self._cache_class.cypher]

    def checksum(self):
        remote_cmd = (f'{self.cypher_command} {self.rpath} | '
                      'awk \'{ print $1 }\';')

        hex_ = self._ssh(remote_cmd).decode(self.encoding)
        log.debug(hex_)
        return bytes.fromhex(hex_)

    def _stat_cmd(self, stat_format=StatResult.stat_format, path=None):
        # TODO use _stat_format_darwin for cases where gstat is missing
        cmd = 'gstat' if self.remote_platform == 'darwin' else 'stat'
        if path is None:
            path = self.rpath

        if path == '':
            _path = path
        else:
            _path = f' "{path}"'

        return f'{cmd} -c {stat_format}{_path}'

    def stat(self):
        remote_cmd = self._stat_cmd()
        out = self._ssh(remote_cmd)
        try:
            return StatResult(out)
        except ValueError as e:
            if out.endswith(b'Permission denied'):
                raise PermissionError(out.decode())

            elif out.endswith(b'No such file or directory'):
                raise FileNotFoundError(out.decode())

            else:
                log.error(remote_cmd)
                raise ValueError(out) from e

    def exists(self):
        try:
            st = self.stat()
            return bool(st)  # FIXME
        except FileNotFoundError:  # FIXME there will be more types here ...
            pass

    @property
    def __parent(self):  # no longer needed since we inherit from path directly
        # because the identifiers are paths if we move
        # file.ext to another folder, we treat it as if it were another file
        # at least for this SshRemote path, if we move a file on our end
        # the we had best update our cache
        # if someone else moves the file on the remote, well, then
        # that file simply vanishes since we weren't notified about it
        # if there is a remote transaction log we can replay if there isn't
        # we have to assume the file was deleted or check all the names and
        # hashes of new files to see if it has moved (and not been changed)
        # a move and change without a sync will be bad for us

        # If you have an unanchored path then resolve()
        # always operates under the assumption that the
        # current working directory which I think is incorrect
        # as soon as you start passing unresolved paths around
        # the remote system doesn't know what context you are in
        # so we need to fail loudly
        # basically force people to manually resolve their paths
        return self.__class__(self.cache.parent)  # FIXME not right ...

    def is_dir(self):
        remote_cmd = self._stat_cmd(stat_format="%F")
        out = self._ssh(remote_cmd)
        return out == b'directory'

    def is_file(self):
        remote_cmd = self._stat_cmd(stat_format="%F")
        out = self._ssh(remote_cmd)
        return out == b'regular file'

    @property
    def children(self):
        # this is amusingly bad, also children_recursive ... drop the maxdepth
        #("find ~/files/blackfynn_local/SPARC\ Consortium -maxdepth 1 "
        #"-exec stat -c \"'%n' %o %s %W %X %Y %Z %g %u %f\" {} \;")
        # chechsums when listing children? maybe ...
        #\"'%n' %o %s %W %X %Y %Z %g %u %f\"
        if self.is_dir():
            # no children if it is a file sadly
            remote_cmd = (f"cd {self.rpath};"
                          f"{self._stat_cmd(path='')} {{.,}}*;"
                          "echo '----';"
                          f"{self.cypher_command} {{.,}}*;"  # FIXME fails on directories destroying alignment
                          'cd "${OLDPWD}"')

            out = self._ssh(remote_cmd)
            stats, checks = out.split(b'\r\n----\r\n')
            #print(stats)
            stats = {sr.name:sr for s in stats.split(b'\r\n')
                     for sr in (StatResult(s),)}
            checks = {fn:bytes.fromhex(cs) for l in checks.split(b'\r\n')
                      if not b'Is a directory' in l
                      for cs, fn in (l.decode(self.encoding).split('  ', 1),)}

            return stats, checks  # TODO

    def _mkdir_child(self, child_name):
        raise NotImplementedError('implement in subclass and/or fix instantiation/existence issues')

    def __repr__(self):
        return f'{self.__class__.__name__}({self.rpath!r}, host={self.host!r})'


SshRemote._bind_flavours()



import zipfile
from pathlib import PurePath
from augpathlib import AugmentedPath
from augpathlib import exceptions as exc
from augpathlib.utils import log as _log


class ZipInternalPath(AugmentedPath):  # should be a pure path, but need bind_flavours

    zip_file = None  # set in _new

    # _paths is basically the fake file system for the contents of the zip file
    _paths = tuple()  # set in _new, and should be a dict +set+ so new files can be added to a zip

    @classmethod
    def _new(cls, zip_path):
        zipath = type('ZipInternalPath',
                    (ZipInternalPath,),
                    dict(zip_file=zipfile.ZipFile(zip_path)))
        zipath._bind_flavours()
        paths = set(zipath(info=zi) for zi in zipath.zip_file.filelist)
        parents = set(parent for path in paths for parent in path.parents)
        zipath._paths = {p:p for p in parents | paths}
        return zipath

    def __new__(cls, *args, info=None, **kwargs):
        if not args and info is not None:
            args = info.filename,

        self = super().__new__(cls, *args, **kwargs)
        if self in cls._paths:
            return cls._paths[self]  # keep only a single copy

        if info is not None:
            self._zi = info

        return self

    @property
    def rchildren(self):
        # FIXME only the actual children not just the whole list
        for path in self._paths:
            if self in path.parents:  # FIXME horrible implementation
                yield path

    def exists(self):
        return self in self._paths

    def is_dir(self):
        return not hasattr(self, '_zi') or self._zi.is_dir()

    def __fspath__(self):
        if self.parent:
            raise TypeError('you need to pass opener= explicitly because python\'s '
                            'builtin open only works on actual files (sigh)')
        else:
            return super().__fspath__()

    def open(self, *args, **kwargs):
        if not self.is_dir():
            self.__fd = self.zip_file.open(self._zi, *args, **kwargs)
            # FIXME probably need a wrapped filed descriptor to clean up the
            # internal state of this class too?
            return self.__fd
        else:
            raise IsADirectoryError(self)

    def close(self):
        # FIXME very bad
        try:
            self.__fd.close()
        except AttributeError:
            raise BaseException('TODO')  # can't close a closed file?


ZipInternalPath._bind_flavours()


class ZipHelper:

    _zfc_cache = {}

    def exists(self):
        return (self.zip_path is None and super().exists() or
                self.zip_relative_path.exists())

    @property
    def ZipInternalPath(self):
        zp = self.zip_path
        if zp is not None and zp not in self._zfc_cache:
            zipath = ZipInternalPath._new(zp)
            self._zfc_cache[zp] = zipath

        return self._zfc_cache.get(zp, None)

    @property
    def zip_relative_path(self):
        if not hasattr(self, '_c_zip_relative_path'):
            self._c_zip_relative_path = None
            zp = self.zip_path
            if zp is not None:
                self._c_zip_relative_path = self.ZipInternalPath(self.relative_to(zp))

        # TODO consider whether we error on not in zip?
        return self._c_zip_relative_path

    path_relative_zip = zip_relative_path

    def asInternal(self):
        return self.zip_relative_path

    @property
    def zip_path(self):
        # like RepoPath.working_dir
        if not hasattr(self, '_c_zip_path'):
            parent = self.parent
            if parent == self:
                self._c_zip_path = None
            else:
                pzp = parent.zip_path  # zip files inside of zip files :/
                if pzp is None and self.suffix == '.zip' or zipfile.is_zipfile(self):
                    self._c_zip_path = self
                    self.ZipInternalPath()
                else:
                    self._c_zip_path = pzp

        return self._c_zip_path

    @property
    def rchildren(self):
        # FIXME internal to external path mangling
        if self.zip_path is not None:
            for child in self.zip_relative_path.rchildren:
                yield self.zip_path / child
        else:
            yield from super().rchildren

    def open(self, *args, **kwargs):
        if self.zip_path is not None:
            return self.path_relative_zip.open(*args, **kwargs)
        else:
            return super().open(*args, **kwargs)
            

class ZipPath(ZipHelper, AugmentedPath):
    pass


ZipPath._bind_flavours()



import os
import re
from setuptools import setup


def find_version(filename):
    _version_re = re.compile(r"__version__ = '(.*)'")
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            return version_match.group(1)


__version__ = find_version('augpathlib/__init__.py')

with open('README.md', 'rt') as f:
    long_description = f.read()

try:
    if os.name == 'nt':
        raise ImportError('the alternate will fail with a TypeError')
    else:
        import magic
        if hasattr(magic, 'libmagic'):
            magic_dep = "python-magic; os_name != 'nt'"
        else:
            magic_dep = "file_magic; os_name != 'nt'"
except ImportError:
    magic_dep = "python-magic; os_name != 'nt'"

tests_require = ['pytest']
setup(
    name='augpathlib',
    version=__version__,
    description='Augmented pathlib. Everything else you could do with a path.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tgbugs/augpathlib',
    author='Tom Gillespie',
    author_email='tgbugs@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],
    keywords='pathlib path paths',
    packages=['augpathlib'],
    python_requires='>=3.6',
    tests_require=tests_require,
    install_requires=[
        'gitpython',
        magic_dep,
        'pexpect>=4.7.0',
        #'psutil',
        'python-dateutil',
        "pyxattr; os_name != 'nt'",
        'terminaltables',
        #'Xlib',
    ],
    extras_require={'dev': ['pytest-cov', 'wheel'],
                    'test': tests_require},
    entry_points={
        'console_scripts': [
        ],
    },
)



import unittest
from augpathlib import exceptions as exc
from .common import TestPathHelper, LocalPathTest, CachePathTest, RemotePathTest
from .common import log


class TestMove(TestPathHelper, unittest.TestCase):
    def _mkpath(self, path, time, is_dir):
        if path.exists():
            return

        if not path.parent.exists():
            yield from self._mkpath(path.parent, time, True)

        if is_dir:
            path.mkdir()
        else:
            path.touch()

        yield path.cache_init(path.metaAtTime(time))

    def _test_move(self, source, target, target_exists=False):
        s = self.test_path / source
        t = self.test_path / target
        #remote = RemotePathTest.invAtTime(1)
        caches = list(self._mkpath(s, 1, int(s.metaAtTime(1).id) in RemotePathTest.dirs))
        if target_exists:  # FIXME and same id vs and different id
            target_caches = list(self._mkpath(t, 2, int(t.metaAtTime(2).id) in RemotePathTest.dirs))

        cache = caches[-1]
        meta = t.metaAtTime(2)
        log.debug(f'{source} -> {target} {cache.meta} {meta}')
        cache.move(target=t, meta=meta)
        assert t.cache.id == RemotePathTest.invAtTime(t, 2)

    def test_0_0_test_cache_local(self):
        c = self.test_path.cache
        assert hasattr(c, '_local_class')
        assert c.local

    def test_0_dir_moved(self):
        source = 'a.e'
        target = 'b.e'
        self._test_move(source, target)

    def test_1_file_moved(self):
        source = 'c.e'
        target = 'd.e'
        self._test_move(source, target)

    def test_2_parent_moved(self):
        source = 'ee/ff/gg'
        target = 'hh/ff/gg'
        self._test_move(source, target)

    def test_3_parents_moved(self):
        source = 'ii/jj/kk'
        target = 'll/mm/kk'
        self._test_move(source, target)

    def test_4_all_moved(self):
        source = 'nn/oo/pp'
        target = 'qq/rr/ss'
        self._test_move(source, target)

    def test_5_onto_self(self):
        source = 't.e'
        target = 't.e'
        self._test_move(source, target)

    def test_6_onto_different(self):
        source = 'a.e'
        target = 't.e'
        try:
            self._test_move(source, target)
            raise AssertionError('should have failed')
        except exc.PathExistsError:
            pass


class TestMoveTargetExists(TestMove):
    def _test_move(self, source, target):
        super()._test_move(source, target, target_exists=True)
        # since all the ids match this should work ...
        #try:
            #raise AssertionError('should have failed')
        #except exc.PathExistsError:
            #pass



import pathlib
from urllib.parse import urlparse
from augpathlib import AugmentedPath
from augpathlib import exceptions as exc
from augpathlib.utils import log as _log

log = _log.getChild('repo')


class RepoHelper:
    _repo_class = None  # set in _setup
    _repos = {}  # repo cache

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    _renew = __new__

    def __new__(cls, *args, **kwargs):
        RepoHelper._setup(*args, **kwargs)
        RepoHelper.__new__ = RepoHelper._renew
        return super().__new__(cls, *args, **kwargs)

    @staticmethod
    def _setup(*args, **kwargs):
        import git
        from git import Repo
        from . import repo_patch
        RepoHelper._git = git
        RepoHelper._repo_class = Repo

    def clone_path(self, remote):
        """ get the path to which a repo would clone
            this makes it possible to check for various issues
            prior to calling init(remote)
        """
        name = pathlib.PurePath(remote).stem
        return self / name

    def clone_from(self, remote, *, depth=None):
        """ clone_from uses the path of the current object as the
            parent path where a new folder will be created with the remote's name

            NOTE: clone_from always uses the remote's naming convention if you want
            to clone into a folder with a different name use init(remote) instead

            NOTE: this does not return the new repo it returns the new
            child path at which the repo is located

            You should probably not use this method since it is poorly designed
            because it requires error handling in the case where a repository
            with the name of the remote has already been cloned as a child of
            the current path
        """
        repo_path = self.clone_path(remote)
        # in a more specific application a variety of tests should go here
        repo_path.init(remote, depth=depth)
        return repo_path

    def init(self, remote=None, depth=None):
        """ NOTE: init conflates init with clone_from
            in cases where a path is known before a remote

            No bare option is provided for init since we assume
            that if you are using this class then you probably
            want the files in the working tree

            NOTE: this does not protect from creating repos
            that contain other repos already, only from creating
            a nested repo inside an existing repo """

        # TODO is_dir() vs is_file()?
        try:
            repo = self.repo
        except exc.NotInRepoError:
            repo = None

        if repo is not None:
            if not self.exists():
                msg = f'how!? {self!r} != {repo.working_dir}'
                assert repo.working_dir == self.as_posix(), msg
                log.warning(f'stale cache on deleted repo {self!r}')
                self._repos.pop(self)
            else:
                raise exc.RepoExistsError(f'{repo}')

        if remote is not None:
            if isinstance(remote, pathlib.Path):
                remote = str(remote)

            repo = self._repo_class.clone_from(remote, self, depth=depth)
        else:
            repo = self._repo_class.init(self)

        self._repos[self] = repo
        return repo

    @property
    def repo(self):
        wd = self.working_dir
        if wd in self._repos:
            return self._repos[wd]
        elif wd is not None:
            repo = self._repo_class(wd.as_posix())
            self._repos[wd] = repo
            return repo
        else:
            raise exc.NotInRepoError(f'{self} is not in a git repository')

    @property
    def working_dir(self):
        # TODO match git behavior here
        # https://github.com/git/git/blob/master/setup.c#L903
        # https://github.com/git/git/blob/08da6496b61341ec45eac36afcc8f94242763468/setup.c#L584
        # https://github.com/git/git/blob/bc12974a897308fd3254cf0cc90319078fe45eea/setup.c#L300
        if (self / '.git').exists():
            if not self.is_absolute():
                # avoid cases where RepoPath('.') gets put in the repos cache
                return self.expanduser().absolute()
            else:
                return self

        elif str(self) == self.anchor:  # anchor is portable
            return None

        else:
            if not self.is_absolute():
                return self.expanduser().absolute().parent.working_dir
            else:
                return self.parent.working_dir

    @property
    def repo_relative_path(self):
        """ working directory relative path """
        repo = self.repo
        if repo is not None:
            if not self.is_absolute():
                path = self.expanduser().absolute()
            else:
                path = self

            return path.relative_to(repo.working_dir)

    path_relative_repo = repo_relative_path

    def _remote_uri(self, prefix, infix=None, ref=None):
        if isinstance(ref, self._git.Commit):
            ref = str(ref)

        repo = self.repo
        remote = repo.remote()
        rnprefix = remote.name + '/'
        url_base = next(remote.urls)
        if url_base.startswith('git@'):
            url_base = 'ssh://' + url_base

        pu = urlparse(url_base)
        netloc = pu.netloc
        path = pu.path
        if netloc.startswith('git@github.com'):
            _, group = netloc.split(':')
            netloc = 'github.com'
            path = '/' + group + path

        if netloc == 'github.com':

            if not ref or ref == 'HEAD':
                ref = repo.active_branch.name
            elif (ref not in [r.name.replace(rnprefix, '') for r in repo.refs] and
                  ref not in [c.hexsha for c in repo.iter_commits(ref, max_count=1)]):
                log.warning(f'unknown ref {ref}')

            if infix is not None:
                rpath = pathlib.PurePosixPath(path).with_suffix('') / infix / ref / self.repo_relative_path
            else:
                rpath = pathlib.PurePosixPath(path).with_suffix('') / ref / self.repo_relative_path
            return prefix + rpath.as_posix()
        elif url_base.startswith('/') and self.__class__(url_base).exists():
            other_rp = self.__class__(url_base)
            other = other_rp / self.repo_relative_path
            return other._remote_uri(prefix, infix, ref)
        else:
            raise NotImplementedError(url_base)

    def remote_uri_human(self, ref=None):
        return self._remote_uri('https://github.com', infix='blob', ref=ref)

    def remote_uri_machine(self, ref=None):
        return self._remote_uri('https://raw.githubusercontent.com', ref=ref)

    def latest_commit(self, rev=None):
        try:
            return next(self.repo.iter_commits(rev=rev,
                                               paths=self.expanduser().resolve().as_posix(),
                                               max_count=1))
        except StopIteration as e:
            raise exc.NoCommitsForFile(self) from e

    def commits(self, *, rev=None, max_count=None):
        yield from self.repo.iter_commits(
            rev=rev,
            paths=self.expanduser().resolve().as_posix(),
            # TODO --follow
            max_count=max_count)

    # a variety of change detection

    def modified(self):
        """ has the filed been changed against index or HEAD """
        return self._do_diff(self.repo.index, None)
        #return self.diff()

    def indexed(self):
        """ cached, or in the index, something like that """
        return self._do_diff(self.repo.head.commit, self.repo.index)
        #return self.diff('HEAD', '')

    def has_uncommitted_changes(self):
        """ indexed or modified aka test working tree against HEAD """
        return self._do_diff(self.repo.head.commit, None)
        #return self.diff('HEAD')

    def _do_diff(self, this, other, *, create_patch=False):
        """ note that the order is inverted from self.diff """
        if not self.exists():
            raise FileNotFoundError(f'{self}')

        list_ = this.diff(other=other, paths=self.repo_relative_path.as_posix(), create_patch=create_patch)
        if list_:
            return list_[0]

    def diff(self, ref='', ref_orig=None, create_patch=False):
        """ ref can be HEAD, branch, commit hash, etc.

            default behaviors diffs the working tree against the index or HEAD if no index

            if ref = None, diff against the working tree
            if ref = '',   diff against the index

            if ref_orig = None, diff from the working tree
            if ref_orig = '',   diff from the index
        """

        def ref_to_object(ref_):
            if ref_ is None:
                return None
            elif ref_ == '':
                return self.repo.index
            else:
                return self.repo.commit(ref_)

        this = ref_to_object(ref_orig)
        other = ref_to_object(ref)

        if this is None:
            if other is None:
                return  # FIXME align return type
            else:
                this, other = other, this

        diff = self._do_diff(this, other, create_patch=create_patch)
        # TODO None -> '' I think?
        # TODO do we render this here or as an extension to the diff?
        return diff

    # commit this file

    def add_index(self):
        """ git add -- {self} """
        # FIXME workaround for broken GitPython behavior on windows
        # C:\\Users\\tom\\git\\repo-name != C:/Users/tom/git/repo-name
        # because they are testing with strings
        # so we use repo_relative_path to avoid the broken
        # _to_relative_path in GitPython
        self.repo.index.add([self.repo_relative_path.as_posix()])

    def commit(self, message, *, date=None):
        """ commit from index
            git commit -m {message} --date {date} -- {self}
        """
        # TODO
        # use a modified Index.write_tree create an in memory tree
        # filtering out changed files that are not the current file
        # during the call to mdb.stream_copy, though it seems like
        # the internal call to write_tree_from_cache may be writing
        # all changes and calculating the sha from that so it may
        # make more sense to try to filter entries instead ...
        # but that means a blob may still be sitting there and
        # get incorporated? I may have to use the full list of entries
        # but sneekily swap out the entries for other changed files for
        # the unmodified entry for their object, will need to experiment


        if self.repo.active_branch.is_valid(): # handle the empty repo case
            # HAH take that complexity!
            _mes = self.repo.git.stash('push')
            try:
                # FIXME concurrent modification?!
                try:
                    self.repo.git.checkout('stash@{0}', '--', self.as_posix())
                except self._git.exc.GitCommandError:
                    pass

                commit = self.repo.index.commit(message=message,
                                                author_date=date)
                return commit
            finally:
                try:
                    self.repo.git.stash('pop')
                except self._git.exc.GitCommandError:
                    pass
        else:
            try:
                self.repo.tree()
                raise ValueError('we are not in the situation we though we were in')
            except:
                commit = self.repo.index.commit(message=message,
                                                author_date=date)
                return commit

    def commit_from_working_tree(self, message, *, date=None):
        """ commit from working tree by automatically adding to index
            git add -- {self}
            git commit -m {message} --date {date} -- {self}
        """
        self.add_index()
        return self.commit(message, date=date)

    # show version file

    def show(self, ref=''):
        # TODO make sure ref='' -> index

        if ref is None:
            with open(self, 'rb') as f:
                return f.read()
        else:
            decoded = self.repo.git.show(ref + ':' + self.repo_relative_path.as_posix())
            return decoded.encode('utf-8', 'surrogateescape')

        return  # unfortunately traversing the tree is a pain for this
        def ref_to_object(ref_):
            if ref_ is None:
                return None
            elif ref_ == '':
                return self.repo.index
            else:
                return self.repo.commit(ref_)

        this = ref_to_object(ref)

        if this is None:
            with open(self, 'rb') as f:
                return f.read()
        else:
            p = self.repo_relative_path.as_posix()
            for blob in this.tree.blobs:
                if blob.path == p:
                    ds = blob.data_stream
                    fcs = ds[-1]
                    return fcs.read()


class RepoPath(RepoHelper, AugmentedPath):
    def rmtree(self, ignore_errors=False, onerror=None, DANGERZONE=False):
        if self in self._repos:
            # remove the reference to the soon to be stale repo
            # in order to prevent GitPython cmd.Git._get_persistent_cmd
            # from causing errors on windows
            # https://github.com/gitpython-developers/GitPython/issues/553
            # https://github.com/gitpython-developers/GitPython/pull/686
            self._repos.pop(self).close()

        super().rmtree(ignore_errors=ignore_errors,
                       onerror=onerror,
                       DANGERZONE=DANGERZONE)


RepoPath._bind_flavours()



import os
import sys
import ctypes

# linux renameat2
# osx renamex_np
# osx exchangedata
# win MoveFileTransacted very few systems support this

def swap_not_implemented(self, target):
    raise NotImplementedError('Windows has not atomic swap operation!')


if sys.platform == 'linux':  # windows will error on ctypes.CDLL without this
    # setup for calling renameat2
    SYS_renameat2 = 316  # from /usr/include/asm/unistd_64.h
    RENAME_EXCHANGE = (1 << 1)  # /usr/src/linux/include/uapi/linux/fs.h
    libc = ctypes.CDLL(None)
    rnat2_syscall = libc.syscall
    rnat2_syscall.restypes = ctypes.c_int       # returns and int
    rnat2_syscall.argtypes = (ctypes.c_int,     # syscall number
                            ctypes.c_int,     # old dir fd
                            ctypes.c_char_p,  # oldpath
                            ctypes.c_int,     # new dir fd
                            ctypes.c_char_p,  # newpath
                            ctypes.c_uint)    # flags

    def swap_linux(self, target):
        """ use renameat2 to perform an atomic swap operation """
        old_fd = os.open(self, 0)
        new_fd = os.open(target, 0)
        old_path = os.fspath(self).encode()
        new_path = os.fspath(target).encode()
        value = rnat2_syscall(SYS_renameat2,
                            old_fd, old_path,
                            new_fd, new_path,
                            RENAME_EXCHANGE)
        os.close(old_fd)
        os.close(new_fd)
        if value != 0:
            raise OSError(value, 'renameat2 failed', self, None, target)
